/****************************************************************************
 * NCSA HDF                                                                 *
 * Software Development Group                                               *
 * National Center for Supercomputing Applications                          *
 * University of Illinois at Urbana-Champaign                               *
 * 605 E. Springfield, Champaign IL 61820                                   *
 *                                                                          *
 * For conditions of distribution and use, see the accompanying             *
 * COPYING file.                                                            *
 *                                                                          *
 ****************************************************************************/
 
/**************************************************************************
  Original code framework came from X-Mosaic.
     Some of the old code exits and should be cleaned up, especially the
  parts dealing with Vgroup/Vdatas, the Image handling code
  and the SDS and attribute handling.
     Bear in mind this whole program should be rewritten.
  This code defintely should NOT be used as an example of *good* 
  programming with HDF...prototypes should ways be thrown out:-)


  This module provides access routines to take an HDF file and return an
  HTML description of its contents.

MAIN MODULES
   sdbGetImage       - extracts selected image from file
   sdbGrokFile       - Process whole HDF/netCDF file and find objects at
                       top level of hierarchy.
   myGetGIFPixel     - returns value of next pixel(used by gif routines)
   sdbGrokRef        - Process HDF/netCDF from a given point in the file
                       i.e. from the reference number. Used primarily
                       ot navigate through a forest of Vgroups.
   sdbGrokImage      - extracts subsampled image from file
  
   sdsGetImage       - convert sds to image
   sdsGrokRef        - get image based on the Query string
   sdsGrokFile       - objects(sds) in the HDF
  
   do_init           - initialize all local variables in global struct

****************************************************************************/

#ifdef RCSID
static char RcsId[] = "@(#)sdb.c,v 1.53 1996/05/17 19:42:23 xlu Exp";
#endif

/* sdb.c,v 1.53 1996/05/17 19:42:23 xlu Exp */

/* Include our stuff */
#define DMAIN_C 
#include "sdb.h"
#include "sdb_util.h"
#include "sdb_ann.h"
#include "sdb_attr.h"
#include "sdb_pal.h"
#include "sdb_ri.h"
#include "sdb_sds.h"
#include "sdb_vd.h"
#include "sdb_vg.h"
#include "sdsdump.h"
#include "show.h"
#include "htmlstrings.h"
#include "html_util.h"

/* inserted by lpd, two new function prototype */
extern char *sds_subset(char *fname, int current_entry, entry entries[MAX_ENTRIES], 
       int num_entries,lvar_st *l_vars, mime_content_type *mime_type);
extern char *vd_subset(char *fname, int32 t, int32 r, lvar_st *l_vars,mime_content_type *mime_type);

/* netCDF error toggle */
IMPORT int ncopts;

/*---------------------- Local variables --------------------------------*/

/* This is not good to handle multiple contexts but until the GIF
   code is retrofitted we can't handle more than one context for now */
LOCAL unsigned char *mypixels;
LOCAL int            myrowlen;


/* ------------------------- MAIN MODULES -------------------------*/


/*--------------------------------------------------------------------
 NAME

 DESCRIPTION
   Initialze all local varibles in global struct /

 RETURNS

--------------------------------------------------------------------*/
void
do_init(lvar_st *l_vars)
{
    l_vars->hfp = NULL;
    l_vars->gfp = NULL;
    l_vars->brief = 0;
    l_vars->my_url = NULL;
    l_vars->hdf_path = NULL;
    l_vars->hdf_path_t = NULL;
    l_vars->f_name     = NULL;
    l_vars->f_path_r   = NULL;
    l_vars->hdf_file   = NULL;
    l_vars->hdf_file_r = NULL;
    l_vars->hdf_ref    = NULL;
    l_vars->display_mode  = 1;
    l_vars->do_netcdf     = 0;
    l_vars->sds_list      = NULL;
    l_vars->gattr_list    = NULL;
    l_vars->dimvar_list   = NULL;
    l_vars->regvar_list   = NULL;
    l_vars->pal_list   = NULL;
    l_vars->rig_list   = NULL;
    l_vars->an_list   = NULL;
    l_vars->flan_list   = NULL;
    l_vars->fdan_list   = NULL;
    l_vars->olan_list   = NULL;
    l_vars->odan_list   = NULL;
    l_vars->vg_list   = NULL;
    l_vars->tvg_list   = NULL;
    l_vars->ovg_list   = NULL;
    l_vars->vd_list   = NULL;
    l_vars->nfields   = 0;
    l_vars->scaling = 0;
    l_vars->field_indices = NULL;
        l_vars->yfield_indices = NULL;
        l_vars->nyfields = 0;
        l_vars->xyplot_name = NULL;
        l_vars->x_axis_name = NULL;
        l_vars->y_axis_name = NULL;
 
    l_vars->vh_start_rec = 0;
    l_vars->vh_end_rec = 0;
    l_vars->plane      = 1;
    DBUG_PRINT(1,(LOGF,"l_vars->plane = %d\n",l_vars->plane ));
    
        /* for IMAGE and SDS */
    l_vars->sdb_ref = 0;
    l_vars->sdb_imgFlag     = 0;
    l_vars->sdb_plane       = 0;
    l_vars->sdb_start_rec   = 0;
    l_vars->sdb_end_rec     = 0;
    /* end of for IMAGE and SDS */
    l_vars->obj_tag = 0;
    l_vars->obj_ref = 0;
    l_vars->sub_s = 0;
    l_vars->sd_dim_start = NULL;
    l_vars->sd_dim_stride = NULL;
    l_vars->sd_dim_end = NULL;
    l_vars->ndims = 0;
} /* do_init () */


/*--------------------------------------------------------------------------
 NAME
       sdbGetImage
 DESCRIPTION
      returns an ImageInfo struct corresponding to the named reference,
      as in sdbGrokRef.  If subsample == TRUE then subsample the 
      image to fit within a hdfImageSize X hdfImageSize box
      Return the backgroud pixel's index in bg only if the file Image
 RETURNS
       return NULL on failure
--------------------------------------------------------------------------*/
ImageInfo *
sdbGetImage(char *filename,   /* HDF/netCDF file */
            char *reference,  /* reference number of image */
            intn subsample,   /* flag: TRUE - fit image within HDFImageSize^2 box */
            intn *bg,         /* background pixel index */
            lvar_st *l_vars)
{
    ImageInfo *Image = NULL;    /* Image info struct */
    int32 tag, ref;      /* tag,ref of image */
    intn  i, j;
    int32 status;
    int32 w, h;          /* image width,height */
    intn  isp;           /* flag: TRUE- there is a palette with image */
    uint8 pal[768];       /* pallette */
    ImageInfo *ret_value = NULL;

    ENTER(2,"sdbGetImage");
    DBUG_PRINT(2,(LOGF,"sdbGetImage: filename=%s,reference=%s \n",
                  filename,reference));

          /* Get tag/ref pair */
          if(sscanf(reference, "tag=%d,ref=%d", &tag, &ref) == 2) 
            {
                switch(tag)
                  {
                  case DFTAG_RIG:
                  case DFTAG_RI8:
                            /* Set read reference number */
                            if((status = DFR8readref(filename, (uint16) ref)) == FAIL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }

                            /* Get dimensions of image */
                            if ((status = DFR8getdims(filename, &w, &h, &isp)) == FAIL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }
                            DBUG_PRINT(1,(LOGF," w=%d,h=%d\n", w,h));

                            /* allocate the basic structure */
                            if ((Image = (ImageInfo *) HDgetspace(sizeof(ImageInfo))) == NULL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }

                            /* allocate space for the data */
                            if ((Image->image_data = (unsigned char *) 
                                 HDgetspace(w * h * sizeof(unsigned char))) == NULL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }

                            /* allocate the palette space */
                            if ((Image->reds = (int *) HDgetspace(256 * sizeof(int))) == NULL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }
                            if ((Image->greens = (int *) HDgetspace(256 * sizeof(int))) == NULL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }
                            if ((Image->blues = (int *) HDgetspace(256 * sizeof(int))) == NULL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }

                            /* fill in the static fields */
                            Image->ismap  = FALSE;
                            Image->width  = w;
                            Image->height = h;
                            Image->num_colors = 256;

                            DBUG_PRINT(1,(LOGF," image->width=%d\n", Image->width));
                            DBUG_PRINT(1,(LOGF," image->height=%d\n", Image->height));

                            /* read the image */
                            if ((status = 
                                 DFR8getimage(filename, Image->image_data, w, h, pal)) == FAIL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }

                            /* set the palette */
                            if(isp) 
                              { /* move the palette over into the fields */
                                  for(i = 0; i < 256; i++) 
                                    {
                                        Image->reds  [i] = (int)pal[i * 3]     ;
                                        Image->greens[i] = (int)pal[i * 3 + 1] ;
                                        Image->blues [i] = (int)pal[i * 3 + 2] ;
                                    }

                                  /* dont mess with palette */
                                  Image->color_map = pal;
                              } 
                            else 
                              { /* create a fake palette */
                                  for(i = 0; i < 256; i++) 
                                    {
                                        Image->reds  [i] = i;
                                        Image->greens[i] = i;
                                        Image->blues [i] = i;
                                    }
                                  /* dont mess with palette */
                                  Image->color_map = NULL;
                              }

                            /* Handle sub-sampling */
                            if(subsample) 
                              {
                                  int max   = (h > w ? h : w);
                                  int skip;   /* skip factor */

                                  DBUG_PRINT(1,(LOGF, "[***] max '%d'\n", max));
                                  DBUG_PRINT(1,(LOGF, "[***] hdfImageSize '%d'\n", hdfImageSize));
                                  DBUG_PRINT(1,(LOGF," we are subsampling\n"));

                                  skip = max / hdfImageSize;

                                  if((skip)&&(w/skip)&&(h/skip)) 
                                    { 
                                        int i, j;
                                        int cnt = 0;
                                        unsigned char * newSpace = NULL;

                                        /* allocate space for the data */
                                        if ((newSpace = (unsigned char *) 
                                             HDgetspace(max * max * sizeof(unsigned char))) 
                                            == NULL)
                                          {
                                              ret_value = NULL;
                                              goto done;
                                          }

                                        /* sub-sample image */
                                        skip++;
                                        for(j = 0; j < h; j += skip)
                                            for(i = 0; i < w; i += skip) 
                                                newSpace[cnt++] = Image->image_data[i + j * w];

                                        /* Free previousl allocated space */
                                        HDfreespace((void *)(Image->image_data));

                                        /* Point to new subsampled image, 
                                         * adjust height and width accordingly */
                                        Image->image_data = newSpace;  
                                        Image->height = h / skip;
                                        Image->width  = w / skip;
                                        if (w % skip) 
                                            Image->width++;
                                    } /* end if "skip" */
                              } /* end if "subsample" */

                            ret_value = Image;
                            goto done;
                      break;

                  case DFTAG_IP8:
                            if ((status = DFPreadref(filename, (uint16) ref)) == FAIL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }

                            /* allocate the basic structure */
                            if ((Image = (ImageInfo *) HDgetspace(sizeof(ImageInfo))) == NULL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }

                            /* allocate space for the data */
                            h = 30;    /* resonable height */
                            if ((Image->image_data = (unsigned char *) 
                                 HDgetspace(768 * h * sizeof(unsigned char))) == NULL)
                              {
                                  ret_value= NULL;
                                  goto done;
                              }

                            /* allocate the palette space */
                            if ((Image->reds = (int *) HDgetspace(256 * sizeof(int))) == NULL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }
                            if ((Image->greens = (int *) HDgetspace(256 * sizeof(int))) == NULL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }
                            if ((Image->blues = (int *) HDgetspace(256 * sizeof(int))) == NULL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }

                            /* fill in the static fields */
                            Image->ismap = FALSE;
                            Image->width = 256;   /* all values */
                            Image->height = h;
                            Image->num_colors = 256;

                            /* read the image */
                            if ((status = DFPgetpal(filename, pal)) == FAIL)
                              {
                                  ret_value = NULL;
                                  goto done;
                              }

                            /* move the palette over into the fields */
                            for(i = 0; i < 256; i++) 
                              {
                                  Image->reds  [i] = (int)pal[i * 3]     ;
                                  Image->greens[i] = (int)pal[i * 3 + 1] ;
                                  Image->blues [i] = (int)pal[i * 3 + 2] ;

                                  /* create h rows for data region */
                                  for(j = 0; j < h; j++) 
                                      Image->image_data[i + j * 256] = i;
                              }

                            Image->color_map = pal;

                            /* at this point we return a palette */
                            ret_value = Image;                 
                      break;
                  default:
                      ret_value = NULL;
                  } /* end switch tag */

            } /* end if "sscanf()" for HDF tag/ref*/

  done:
    return ret_value;
    EXIT(2,"sdbGetImage");
} /* sdbGetImage */

/*---------------------------------------------------------------------------
 NAME
       sdbGrokFile
 DESCRIPTION
        The function sdbGrokFile() will take a whole file and return stuff that
        can be seen at the "top level" (i.e. all datasets, all raster images 
        and all Vgroups which are not contained within other Vgroups).
 RETURNS
----------------------------------------------------------------------------*/
char *
sdbGrokFile(char *fname, 
            entry entries[MAX_ENTRIES], 
            mime_content_type *mime_type,
            int current_entry,
            int num_entries,
            lvar_st *l_vars)
{
    char *tmp = (char *)tempnam(NULL,"html");  /* Temporary file name for HTML desc */
    int32 fid = FAIL;                        /* File handle */
    int32 ret;
    char *ret_value = NULL;

    ENTER(2,"sdbGrokFile");
    DBUG_PRINT(3,(LOGF,"l_vars->do_netcdf=%d \n",l_vars->do_netcdf));
    /* make sure we don't crash on invalid open */
    ncopts = 0;

          /* Open temproary file to write HTML description of HDF/netCDF file */
          if (!(l_vars->hfp = fopen(tmp, "w")))
            {
                ret_value = NULL;
                goto done;
            }
        
          /* Write MIME header */
          *mime_type = TEXT_HTML;
          if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL)
            {
                gateway_err(l_vars->hfp,"sdbGrokFile: writing HTML header",0);
                ret_value = NULL;
                goto done;
            }

    /* are we dumping  HDF/netCDF file */
          if (!l_vars->do_netcdf)
            {
                /* open file once */
                if ((fid = Hopen(fname, DFACC_RDONLY, 0)) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokFile: Problem opening HDF file",0);
                      ret_value = NULL;
                      goto done;
                  }


                /* process all annotations */
                if (do_anns(fname,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokFile: error processing annotations ",0);
                      ret_value = NULL;
                      goto done;
                  }
            }

          /* process SDS in file */
          if (do_sds(fname, l_vars) == FAIL)
            {
                gateway_err(l_vars->hfp,"sdbGrokFile: error processing SDS ",0);
                ret_value = NULL;
                goto done;
            }

          if (!l_vars->do_netcdf)
            {
                /* process pals and rigs in file */
                if (do_pals(fname,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokFile: error processing palettes ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* process raster8 */
                if (do_rigs(fname,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokFile: error processing raster8 ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* process Vgroups */
                if (do_vgs(fname, l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokFile: error processing Vgroups ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* process lone Vdatas */
                if (do_lone_vds(fname,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokFile: error processing lone Vdatas ",0);
                      ret_value = NULL;
                      goto done;
                  }

            } /* end if not netCDF */


          /* 
           * Now dumping time 
           */
          if(!l_vars->do_netcdf)
            {
                DBUG_PRINT(3,(LOGF,"doing annotations \n"));
                /* dump file annotations */
                dump_anns(fname,l_vars);
            }

          /* Do SDSs */
                DBUG_PRINT(3,(LOGF,"doing sdss \n"));
                dump_sds(fname, current_entry, entries, num_entries,l_vars);

          if(!l_vars->do_netcdf)
            {
                /* Do RIGs */
                DBUG_PRINT(3,(LOGF,"doing rigs \n"));
                dump_rigs(fname,l_vars);
        
                DBUG_PRINT(3,(LOGF,"doing lone pals \n"));
                /* Do lone Pals */
                dump_pals(fname,l_vars);
        
                DBUG_PRINT(3,(LOGF,"doing vgroups \n"));
                /* Do Vgroups */
                dump_vgs(fname, l_vars);
        
                DBUG_PRINT(3,(LOGF,"doing lone Vdatas \n"));
                /* Do lone Vdatas */
                dump_lone_vds(fname,l_vars);

                      /* Display graphical representation of Vgroups/Vdatas */
                      fprintf(l_vars->hfp,"<PRE WIDTH=\"%d\">",DOC_WIDTH);
                      ret = show_tree(fname,l_vars);
                      print_label(l_vars->hfp, "</PRE>"); 

            } /* end if not netCDF */

    /* Close temporary file */
        fclose(l_vars->hfp);

    ret_value = tmp;
  done:
    if (!l_vars->do_netcdf )
      {
          /* close file */
          if (fid != NULL)
              Hclose(fid);
      }
    EXIT(2,"sdbGrokFile");
    return ret_value;
} /* sdbGrokFile */

/*-------------------------------------------------------------------- 
 NAME
     myGetGIFPixel
 DESCRIPTION
     Gets value of next pixel
 RETURNS
     returns next pixel value
--------------------------------------------------------------------*/ 
int
myGetGIFPixel (int x, int y)
{
    return (mypixels[y * myrowlen + x]);
} /* myGetGIFPixel */


/*----------------------------------------------------------------------------
 NAME
      sdbGrokRef
 DESCRIPTION
        The function sdbGrokRef() will return a description of what can
        be "seen" from a given location in the file.  Typically, this will be 
        how users can navigate their way through a forest of Vgroups.
 RETURNS
      Returns the file name containg the HTML description or the GIF image
      if successful else NULL.
----------------------------------------------------------------------------*/
char *
sdbGrokRef(char *fname,                 /* HDF/netCDF file name */
           char *ref,                  /* Reference number to start navigation from */
           entry entries[MAX_ENTRIES], 
           mime_content_type *mime_type, 
           int current_entry, 
           int num_entries,
           lvar_st *l_vars)
{
    int32 t, r;           /* tag/ref numbers */
    intn sub_sample;
    char *tmp = (char *)tempnam(NULL,"html");  /* Temporary file name for HTML desc */
    char *gtmp = (char *)tempnam(NULL,"img");  /* Temporary file name for HTML desc */
    char *ptmp = (char *)tempnam(NULL,"pal");  /* Temporary file name for HTML desc */
    int32 fid = FAIL;
    ImageInfo *image;
    intn bkg = 0;
    int i;
    static skip=FALSE;
    char *ret_value = NULL;

    ENTER(2,"sdbGrokRef");
    DBUG_PRINT(1,(LOGF," ref=%s\n", ref));

    /* skip past seperator */
    if (ref[0]==';')
        ref++;

    if(sscanf(ref, "tag=%d,ref=%d,s=%d", &t, &r, &sub_sample) == 3) 
      {
          DBUG_PRINT(1,(LOGF," got tag=%d, ref=%d\n", t,r));
          l_vars->obj_tag = (uint16) t;
          l_vars->obj_ref = (uint16) r;
          l_vars->sub_s = sub_sample;

          switch((uint16) t) 
            {
            case DFTAG_VG:  /* Vgroup */
                DBUG_PRINT(1,(LOGF," tag is Vgroup\n"));
                /* Open temporary file name for HTML description */
                if ((l_vars->hfp = fopen(tmp, "w")) == NULL) 
                  {
                      gateway_err(stdout,"sdbGrokRef: Problem opening temproray file",1);
                      ret_value = NULL;
                      goto done;
                  }
              
                /* Write MIME header response */
                *mime_type = TEXT_HTML;
                if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: writing HTML header",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* open file once */
                if ((fid = Hopen(fname, DFACC_RDONLY, 0)) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: Problem opening HDF file",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* process all annotations */
                if (do_anns(fname,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: error processing annotations ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* process SDS in file */
                if (do_sds(fname, l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"hdfGrokRef: error processing SDS ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* Process Vgroups in file */
                if ((do_vgs(fname, l_vars)) == FAIL)              
                  {
                      gateway_err(l_vars->hfp,"hdfGrokRef: error processing Vgroups ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* Dump Vgroup in file */
                if ((dump_vg(fname, r,l_vars)) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"hdfGrokRef: error dumping Vgroup ",0);
                      ret_value = NULL;
                      goto done;
                  }

                print_label(l_vars->hfp, "<HR>\n");

                break;
            case DFTAG_VH: /* Vdata */
                DBUG_PRINT(1,(LOGF," tag is Vdata\n"));

                /* inserted by lpd */
                if(l_vars->display_mode == 3){	/* add the all-fields option */
                     if(l_vars->nfields == 0 || l_vars->field_indices == NULL){
                           gateway_err(l_vars->hfp, "do_vd: No field is selected<p>You have to select fields to dump\n", 1);
                           ret_value = NULL;
                           goto done;
                     }
                     if(l_vars->view_type == V_IMAGE && ( l_vars->nyfields == 0
                        || l_vars->yfield_indices == NULL)){
                         gateway_err(l_vars->hfp, "Error, you have to select a variable for Y axis for plotting\n", 1);
                     }
 
                     /* check to see if user wants all fields */
                     for(i=0; i<l_vars->nfields; i++){
                         if(l_vars->field_indices[i] == 0)break;
                     }
                     
                     if(i != l_vars->nfields){    /* all fields */
                         l_vars->nfields = l_vars->ntfields;
                         for(i=1; i<=l_vars->nfields; i++)
                              l_vars->field_indices[i-1] = i;
                     }
                }
                           
                if(l_vars->display_mode == 3 && l_vars->view_type != 5){  /* subsetting--lpd */
                    ret_value = vd_subset( fname,t,r,l_vars, mime_type);
                    goto done;
                }
                else {
                    if ((l_vars->hfp = fopen(tmp, "w")) == NULL) 
                    {
                         gateway_err(stdout,"sdbGrokRef: Problem opening temproray file",1);
                         ret_value = NULL;
                         goto done;
                    }

                   /* Write MIME header response */
                   *mime_type = TEXT_HTML;
                   if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL)
                   {
                        gateway_err(l_vars->hfp,"sdbGrokRef: wring HTML headr",0);
                        ret_value = NULL;
                        goto done;
                   }

                   /* process Vdata */
                   if ((do_vd(fname,t,r,l_vars)) == FAIL)
                   {
                        gateway_err(l_vars->hfp,"hdfGrokRef: error dumping Vdata ",0);
                        ret_value = NULL;
                        goto done;
                   }
                }

                break;

            case DFTAG_NDG:
            case 6666: /* fake take for netCDF */
                DBUG_PRINT(1,(LOGF," tag is NDG\n"));

                /* Open temporary file name for HTML description */
                if ((l_vars->hfp = fopen(tmp, "w")) == NULL)
                  {
                      gateway_err(stdout,"sdbGrokRef: Problem opening temproray file",1);
                      ret_value = NULL;
                      goto done;
                  }
              
                /* Write MIME header response */
                *mime_type = TEXT_HTML;
                if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: wring HTML headr",0);
                      fclose(l_vars->hfp);
                      ret_value = tmp;
                      goto done;
                  }

                if (!l_vars->do_netcdf)
                {
                      /* process SDS, inserted by lpd  */ 
                      if(l_vars->display_mode == 3){
                              ret_value = sds_subset(fname, current_entry, entries, num_entries,l_vars, mime_type);
                              if(ret_value != NULL){
                                     fclose(l_vars->hfp);
                                     remove(tmp); 
                                     HDfree(gtmp);
                                     HDfree(ptmp);
                                     HDfree(tmp);
                                     return(ret_value);
                             }
                      }  /* handle the view type: image, ASCII, HDF and binary files. End the insert */
                 

                      /* process all annotations */
                      if (do_anns(fname,l_vars) == FAIL)
                        {
                            gateway_err(l_vars->hfp,"sdbGrokRef: error processing annotations ",0);
                            ret_value = NULL;
                            goto done;
                        }
                  }

                /* process SDS in file */
                if (do_sds(fname, l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"hdfGrokRef: error processing SDS ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* print SDS info */
                if (dump_sd(fname, r,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: unable to access dataset",0);
                      ret_value = NULL;
                      goto done;
                  }

                break;
            case DFTAG_IP8:
                DBUG_PRINT(1,(LOGF," tag is raster palette\n"));

                /* Open temporary file name for HTML description */
                if ((l_vars->hfp = fopen(tmp, "w")) == NULL)
                  {
                      gateway_err(stdout,"sdbGrokRef: Problem opening temproray file",1);
                      ret_value = NULL;
                      goto done;
                  }
              
                /* Write MIME header response */
                *mime_type = IMAGE_GIF;
                if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: writing HTML header",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* open file once */
                if ((fid = Hopen(fname, DFACC_RDONLY, 0)) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"hdfGrokRef: Problem opening HDF file",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* process all annotations */
                if (do_anns(fname,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: error processing annotations ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* process lone pals in file */
                if (do_pals(fname,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"hdfGrokRef: error processing palettes ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* process rigs in file */
                if (do_rigs(fname,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"hdfGrokRef: error processing raster8 ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* Open temporary file name for gif description */
                if ((l_vars->gfp = fopen(ptmp, "wb+")) == NULL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: Problem opening temproray file for GIF",1);
                      ret_value = NULL;
                      goto done;
                  }
              
                /* Get Palette */
                if ((image = sdbGetImage(fname, ref, FALSE, &bkg,l_vars)) == NULL)
                  {
                      fclose(l_vars->gfp);
                      gateway_err(l_vars->hfp,"sdbGrokRef: Problem with sdbGetImage",1);
                      ret_value = NULL;
                      goto done;
                  }

                DBUG_PRINT(1,(LOGF," sdbGetImage returned\n"));
                DBUG_PRINT(1,(LOGF," image->width=%d\n",image->width));
                DBUG_PRINT(1,(LOGF," image->height=%d\n",image->height));
                DBUG_PRINT(1,(LOGF," bkg=%d\n",bkg));

                mypixels = image->image_data;
                myrowlen =image->width;
                DBUG_PRINT(1,(LOGF,"calling GIFEncode \n"));

                GIFEncode(
                    l_vars->gfp,                       /* file to write to */
                    image->width,		/* width */
                    image->height,		/* height */
                    0,			/* No interlacing */
                    bkg,			/* Index of Backgrounf */
                    8,			/* Bits Per pixel */
                    image->reds,		/* Red colormap */
                    image->greens,		/* Green colormap */
                    image->blues,		/* Blue colormap */
                    myGetGIFPixel);		/* Get Pixel value */
                DBUG_PRINT(1,(LOGF,"GIFEncode returned\n"));

                /* Close temporary gif file and html file*/
                fclose(l_vars->gfp);
                fclose(l_vars->hfp);

                /* Free memory allocated in sdbGetImage */
                if (image->image_data != NULL)
                    HDfree(image->image_data);
                if (image->reds != NULL)
                    HDfree(image->reds);
                if (image->greens != NULL)
                    HDfree(image->greens);
                if (image->blues != NULL)
                    HDfree(image->blues);
                if (image != NULL)
                    HDfree(image);

                ret_value = ptmp; /* return palette file name */
                goto done;

                break;
            case DFTAG_RIG:
                DBUG_PRINT(1,(LOGF," tag is raster group\n"));

                /* Open temporary file name for HTML description */
                if ((l_vars->hfp = fopen(tmp, "w")) == NULL)
                  {
                      gateway_err(stdout,"sdbGrokRef: Problem opening temproray file",1);
                      ret_value = NULL;
                      goto done;
                  }

                      /* Open temporary file name for gif description */
                      if ((l_vars->gfp = fopen(gtmp, "wb+")) == NULL)
                        {
                            gateway_err(l_vars->hfp,"sdbGrokRef: Problem opening temproray file for GIF",1);
                            ret_value = NULL;
                            goto done;
                        }

                      /* Write MIME header response */
                      *mime_type = IMAGE_GIF;
                      if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL)
                        {
                            fclose(l_vars->gfp);
                            gateway_err(l_vars->hfp,"sdbGrokRef: writing HTML headr",0);
                            ret_value = NULL;
                            goto done;
                        }

                      /* open file once */
                      if ((fid = Hopen(fname, DFACC_RDONLY, 0)) == FAIL)
                        {
                            gateway_err(l_vars->hfp,"hdfGrokRef: Problem opening HDF file",0);
                            ret_value = NULL;
                            goto done;
                        }

                      /* process all annotations */
                      if (do_anns(fname,l_vars) == FAIL)
                        {
                            gateway_err(l_vars->hfp,"sdbGrokRef: error processing annotations ",0);
                            ret_value = NULL;
                            goto done;
                        }

                      /* process lone pals in file */
                      if (do_pals(fname,l_vars) == FAIL)
                        {
                            gateway_err(l_vars->hfp,"hdfGrokRef: error processing palettes ",0);
                            ret_value = NULL;
                            goto done;
                        }

                      /* process rigs in file */
                      if (do_rigs(fname,l_vars) == FAIL)
                        {
                            gateway_err(l_vars->hfp,"hdfGrokRef: error processing raster8 ",0);
                            ret_value = NULL;
                            goto done;
                        }

                      /* Get image */
                      if ((image = sdbGetImage(fname, ref, sub_sample, &bkg,l_vars)) == NULL)
                        {
                            fclose(l_vars->gfp);
                            gateway_err(l_vars->hfp,"sdbGrokRef: Problem with sdbGetImage",1);
                            ret_value = NULL;
                            goto done;
                        }

                      DBUG_PRINT(1,(LOGF," sdbGetImage returned\n"));
                      DBUG_PRINT(1,(LOGF," image->width=%d\n",image->width));
                      DBUG_PRINT(1,(LOGF," image->height=%d\n",image->height));
                      DBUG_PRINT(1,(LOGF," bkg=%d\n",bkg));

                      mypixels = image->image_data;
                      myrowlen = image->width;
                      DBUG_PRINT(1,(LOGF," myrowlen=%d\n",myrowlen));

                      DBUG_PRINT(1,(LOGF,"calling GIFEncode \n"));

                      GIFEncode(
                          l_vars->gfp,                       /* file to write to */
                          image->width,		/* width */
                          image->height,		/* height */
                          0,			/* No interlacing */
                          bkg,			/* Index of Backgrounf */
                          8,			/* Bits Per pixel */
                          image->reds,		/* Red colormap */
                          image->greens,		/* Green colormap */
                          image->blues,		/* Blue colormap */
                          myGetGIFPixel);		/* Get Pixel value */
                      DBUG_PRINT(1,(LOGF,"GIFEncode returned\n"));

                      /* Close temporary html file and gif file */
                      fclose(l_vars->gfp);
                      fclose(l_vars->hfp);

                      /* Free memory allocated in sdbGetImage */
                      if (image->image_data != NULL)
                          HDfree(image->image_data);
                      if (image->reds != NULL)
                          HDfree(image->reds);
                      if (image->greens != NULL)
                          HDfree(image->greens);
                      if (image->blues != NULL)
                          HDfree(image->blues);
                      if (image != NULL)
                          HDfree(image);

                      ret_value = gtmp;  /* return image file name */
                      goto done;

                break;
            case DFTAG_FID:
            case DFTAG_FD:
            case DFTAG_DIL:
            case DFTAG_DIA:
                DBUG_PRINT(1,(LOGF," tag is an annotation \n"));
                /* Open temporary file name for HTML description */
                if ((l_vars->hfp = fopen(tmp, "w")) == NULL)
                  {
                      gateway_err(stdout,"sdbGrokRef: Problem opening temproray file",1);
                      ret_value = NULL;
                      goto done;
                  }
              
                /* Write MIME header response */
                *mime_type = TEXT_HTML;
                if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: writing HTML headr",0);
                      ret_value = NULL;
                      goto done;
                  }
                /* open file once */
                if ((fid = Hopen(fname, DFACC_RDONLY, 0)) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"hdfGrokRef: Problem opening HDF file",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* process all annotations */
                if (do_anns(fname,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: error processing annotations ",0);
                      ret_value = NULL;
                      goto done;
                  }

                /* print annotation */
                print_ann(fid,r,t,sub_sample,ANtag2atype(t),l_vars->hfp,l_vars);

                /* clenup */
                break;
            default:
                /* Open temporary file name for HTML description */
                if ((l_vars->hfp = fopen(tmp, "w")) == NULL)
                  {
                      gateway_err(stdout,"sdbGrokRef: Problem opening temproray file",1);
                      ret_value = NULL;
                      goto done;
                  }
              
                /* Write MIME header response */
                *mime_type = TEXT_HTML;
                if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL)
                  {
                      gateway_err(l_vars->hfp,"sdbGrokRef: writing HTML header",0);
                      ret_value =  NULL;
                      goto done;
                  }

                /* Indicate we don't handle these objects */
                print_label(l_vars->hfp, ERR1);
                fprintf(l_vars->hfp,"%s %s", 
                        HDgettagsname((uint16)t), ERR2);
                break;
            } /* end switch "t" */
      } 
    else  /* not able to get tag/ref */
      {
          /* Open temporary file name for HTML description */
          if ((l_vars->hfp = fopen(tmp, "w")) == NULL)
            {
                gateway_err(stdout,"sdbGrokRef: Problem opening temproray file",1);
                ret_value = NULL;
                goto done;
            }
              

          /* Write MIME header response */
          *mime_type = TEXT_HTML;
          if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL)
            {
                gateway_err(l_vars->hfp,"sdbGrokRef: writing HTML header",0);
                ret_value = NULL;
                goto done;
            }

          DBUG_PRINT(1,(LOGF," couldn't get tag/ref\n"));

          print_label(l_vars->hfp, ERR1);
      } /* end else not able to get tag/ref */       
   

    /* Check for funky sds handling */
    if ((ref[0]=='#')||(ref[0]=='@'))
      {
          DBUG_PRINT(1,(LOGF,"this is a funky SDS, ref=%s\n",ref));
          dump_sds(fname, current_entry, entries, num_entries,l_vars);
      }

    /* Close temporary file and return its name */
    fclose(l_vars->hfp);

    ret_value = tmp; /* html response file */

  done:
    if (ret_value == NULL)
      {
      }

    /* cleanup */
    if (!l_vars->do_netcdf)
      {
          /* close file */
          if (fid != FAIL)
              Hclose(fid);
      }

    EXIT(2,"sdbGrokRef");
    return ret_value;
} /* sdbGrokRef */


/*---------------------------------------------------------------------------
 NAME
       sdbGrokImage
 DESCRIPTION
       returns an ImageInfo struct corresponding to the named reference,
       as in sdbGrokRef. Calls "sdbGetImage()" with "subsample = TRUE".
 RETURNS
       return NULL on failure
----------------------------------------------------------------------------*/
ImageInfo *
sdbGrokImage(char *filename,    /* HDF/netCDF file */
             char *reference,   /* Reference number of image */
             int *bg,           /* background pixel index */
             lvar_st *l_vars)
{
    /* Return a subsampled image */
    return (sdbGetImage(filename, reference, TRUE, bg,l_vars));
} /* sdbGrokImage */

/* =========================================================  */

/* for sds */

/*--------------------------------------------------------------------------
 NAME
       sdsGetImage
 DESCRIPTION
      returns an ImageInfo struct corresponding to the named ref number,
      as sdsGrokRef.  If subsample == TRUE then subsample the 
      image to fit within a hdfImageSize .
      Return the backgroud pixel's index in bg only if the file Image
 RETURNS
       return NULL on failure
--------------------------------------------------------------------------*/
ImageInfo *
sdsGetImage(char *fname, 
            char *hdu,       /* hdu number of image */
            int subsample,   /* flag: TRUE - fit image within hdfImageSize^2 box */
            int *bg,         /* background pixel index */
            lvar_st *l_vars)
{
    ImageInfo *Image = NULL;    /* Image info struct */
    int32 ref;                  /* tag,ref of image */
    int   plane;                /* plane nuber */
    int   s;
    int32 rank;          /* rankof data */
    intn  i;
    int32 status;
    int32 w, h;          /* image width,height */
    int   numPlane = 1;
    int32 sdid, sdsid;
    char  sdsName[25];
    int32 dimSize[20];
    int32 nt, nattrs;

    ENTER(2,"sdsGetImage");
    
    /* Get ref pair */
    if(sscanf(hdu, "ref=%d,s=%d,plane=%d",&ref,&s,&plane) == 3)  {

        /* Open file for reading */
        if ((sdid = SDstart(fname, DFACC_RDONLY)) == FAIL) {      
            return NULL;
        }

        /* get index for the specified dataset by the ref. */
        if (l_vars->do_netcdf)
          {
            i = ref; /* for netCDF we use the index to indicate ref
                      * as they really don't have NDG */
          }
        else
          {
              i = SDreftoindex(sdid, ref);
              if (i == FAIL)
                  return NULL;
          }
 
        /* select dataset */
        sdsid = SDselect(sdid, i);
        if (sdsid == FAIL) {
            return NULL;
        }
     
        /* Get dimensions of image */
        if ((status=SDgetinfo(sdsid,sdsName,&rank,dimSize,&nt,&nattrs))==FAIL)
            return NULL;
	
        if (rank <2)
            return NULL;

        w = dimSize[rank-1];
        h = dimSize[rank-2];
        for (i=0; i<rank-2; i++)
            numPlane *= dimSize[i];
	

        DBUG_PRINT(1,(LOGF," w=%d,h=%d\n", w,h));
        DBUG_PRINT(1,(LOGF," PLANE=%d\n", plane));

        /* allocate the basic structure */
        if ((Image = (ImageInfo *) HDmalloc(sizeof(ImageInfo))) == NULL)
            return NULL;

        /* allocate space for the data */
        if ((Image->image_data = (unsigned char *) 
             malloc(w * h * sizeof(unsigned char))) == NULL)
            return NULL;

        /* allocate the palette space */
        if ((Image->reds =   (int *)malloc(256 * sizeof(int))) == NULL)
            return NULL;
        if ((Image->greens = (int *)malloc(256 * sizeof(int))) == NULL)
            return NULL;
        if ((Image->blues =  (int *)malloc(256 * sizeof(int))) == NULL)
            return NULL;

        /* fill in the static fields */
        Image->ismap  = FALSE;
        Image->width  = w;
        Image->height = h;
        Image->num_colors = 256;

        DBUG_PRINT(1,(LOGF," image->width=%d\n", Image->width));
        DBUG_PRINT(1,(LOGF," image->height=%d\n", Image->height));

        /* read the image */
        if ((status = 
             sdsImage(sdsid, Image->image_data, plane)) == FAIL)
            return NULL;
	    
        /* set the palette */

        /* create a fake palette */
        for(i = 0; i < 256; i++)  {
		
            Image->reds  [i] = i;
            Image->greens[i] = i;
            Image->blues [i] = i;
        }
        Image->color_map = NULL;

        /* reset the pallete (test only) */
        /* setPallete("image.pal",Image->reds,Image->greens,Image->blues); */

        /* Handle sub-sampling */
        if(subsample) {
              
            int max   = (h > w ? h : w);
            int skip;   /* skip factor */
	    
            DBUG_PRINT(1,(LOGF, "[***] max '%d'\n", max));
            DBUG_PRINT(1,(LOGF, "[***] hdfImageSize '%d'\n", hdfImageSize));
            DBUG_PRINT(1,(LOGF," we are subsampling\n"));

            skip = max / hdfImageSize;

            if((skip)&&(w/skip)&&(h/skip))  {
              
                int i, j;
                int cnt = 0;
                unsigned char * newSpace = NULL;
		
                if (newSpace)
                    HDfree((unsigned char *)newSpace);

                /* allocate space for the data */
                if ((newSpace = (unsigned char *) \
                     HDmalloc(max * max * sizeof(unsigned char))) == NULL)
                    return NULL;

                /* sub-sample image */
                skip++;
                for(j = 0; j < h; j += skip)
                    for(i = 0; i < w; i += skip) 
                        newSpace[cnt++] = Image->image_data[i + j * w];

                /* Free previousl allocated space */
                HDfree((void *)(Image->image_data));

                /* Point to new subsampled image, 
                 * adjust height and width accordingly */
                Image->image_data = newSpace;  
                Image->height = h / skip;
                Image->width  = w / skip;
                if (w % skip) 
                    Image->width++;
            } /* end if "skip" */
        } /* end if "subsample" */

        /* Terminate access to the array  dataset  */
        SDendaccess(sdsid);

        /* Terminate access to the SD interface and close the file */
        SDend(sdid);

        EXIT(2,"sdsGetImage");
        return Image;
    } /* end if "sscanf()" */
    EXIT(2,"sdsGetImage");
} /* sdsGetImage */


/*----------------------------------------------------------------------------
 NAME
      sdsGrokRef
 DESCRIPTION
        The function sdsGrokRef() will return a description of what can
        be "seen" from a given location in the file. 
 RETURNS
      Returns the file name containg the HTML description or the GIF image
      if successful else NULL.
----------------------------------------------------------------------------*/
char *
sdsGrokRef(char *fname, 
           char *ref,
           entry entries[MAX_ENTRIES], 
           mime_content_type *mime_type, 
           int current_entry, 
           int num_entries,
           lvar_st *l_vars)
{
    int32 r;           /* tag/ref numbers */
    int32 x, y;
    int   sub_sample;
    /* Temporary file name for HTML desc */
    char *tmp = (char *)tempnam(NULL,"html"); 
    
    /* Temporary file name for HTML desc */
    char *gtmp = (char *)tempnam(NULL,"img"); 
 
    /* Temporary file name for HTML desc */
    char *ptmp = (char *)tempnam(NULL,"pal");  

    ImageInfo *image;
    int   bkg = 0;
    static skip=FALSE;


    ENTER(2,"sdsGrokRef");
    DBUG_PRINT(1,(LOGF," ref=%s\n", ref));

    if (ref[0]==';')
        ref++;
    if(sscanf(ref, "ref=%d,s=%d", &r, &sub_sample) == 2) {
      
        DBUG_PRINT(1,(LOGF," got ref=%d\n",r));     
        /* Open temporary file name for HTML description */
        if ((l_vars->hfp = fopen(tmp, "w")) == NULL)
            return(NULL);
	
        /* Open temporary file name for gif description */
        if ((l_vars->gfp = fopen(gtmp, "wb+")) == NULL)
            return(NULL);
              
        /* Write MIME header response */
        *mime_type = IMAGE_GIF;
        if (!skip) {                
            if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL) {   
                gateway_err(l_vars->hfp," writing HTML header",0);
                return NULL;
            }
            skip = !(skip);
        }

        /* Get Image */
        if ((image = sdsGetImage(fname,ref,sub_sample,&bkg,l_vars))==NULL) {
            /* Close temporary gif file and html file*/
            fclose(l_vars->gfp);
            fclose(l_vars->hfp);
            return NULL;
        }
        DBUG_PRINT(1,(LOGF," sdsGetImage returned\n"));
        DBUG_PRINT(1,(LOGF," image->width=%d\n",image->width));
        DBUG_PRINT(1,(LOGF," image->height=%d\n",image->height));
        DBUG_PRINT(1,(LOGF," bkg=%d\n",bkg));

        mypixels = image->image_data;
        myrowlen = image->width;
        DBUG_PRINT(1,(LOGF,"calling GIFEncode \n"));

        GIFEncode(
            l_vars->gfp,                       /* file to write to */
            image->width,		/* width */
            image->height,		/* height */
            0,			/* No interlacing */
            bkg,			/* Index of Backgrounf */
            8,			/* Bits Per pixel */
            image->reds,		/* Red colormap */
            image->greens,		/* Green colormap */
            image->blues,		/* Blue colormap */
            myGetGIFPixel);		/* Get Pixel value */

        DBUG_PRINT(1,(LOGF,"GIFEncode returned\n"));

        /* Close temporary gif file and html file*/
        fclose(l_vars->gfp);
        fclose(l_vars->hfp);

        /* Free memory allocated in sdsGetImage */

        if (image->blues != NULL)
            HDfree(image->blues); 

        if (image->greens != NULL)
            HDfree(image->greens); 

        if (image->reds != NULL)
            HDfree(image->reds); 

        if (image->image_data != NULL)
            HDfree(image->image_data);

        if (image != NULL)
            HDfree(image);
   	    
        EXIT(2,"sdsGrokRef");
	/* 
        HDfree(tmp);
        HDfree(ptmp); */

        return(gtmp); /* return image file name */
    }

    else  { /* not able to get tag/ref */
      
        /* Open temporary file name for HTML description */
        if ((l_vars->hfp = fopen(tmp, "w")) == NULL)
            return(NULL);

        /* Write MIME header response */
        *mime_type = TEXT_HTML;
        if (!skip) {
          
            if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL) {
              
                gateway_err(l_vars->hfp," writing HTML header",0);
                return NULL;
            }

            skip = !(skip);
        }

        DBUG_PRINT(1,(LOGF," couldn't get tag/ref\n"));

        if(sscanf(ref, "fileImageHit?%d,%d", &x, &y) == 2) {
          
            char *buf;
            print_label(l_vars->hfp, SDB2);
            print_label(l_vars->hfp, ERR1);

            /* hmmm....not sure */
            buf = malloc(100);
            sprintf(buf, "#DataSet%d", 2);
        } 
        else if ((ref[0]!='#') && (ref[0]!='@')) {
          
            print_label(l_vars->hfp, ERR1);
        }
    } /* end else not able to get tag/ref */       
   
    DBUG_PRINT(1,(LOGF," lets see if this is an SDS, ref=%s\n",ref));
  
    /* Close temporary file and return its name */
    fclose(l_vars->hfp);
   
    EXIT(2,"sdsGrokRef");
    return(tmp);
} /* sdsGrokRef */



/*---------------------------------------------------------------------------
 NAME
       sdsGrokFile
 DESCRIPTION
        The function sdsGrokFile() will just take an SDS within the HDF file 
 RETURNS
        HTML file
----------------------------------------------------------------------------*/
char *
sdsGrokFile(char *fname, 
            entry entries[MAX_ENTRIES], 
            mime_content_type *mime_type,
            int current_entry,
            int num_entries,
            lvar_st *l_vars)
{
    char *tmp = (char *)tempnam(NULL,"html");  /* Temporary file name for HTML desc */
    int32 ref;
    int32 sdid, sdsid;
    int32 i;
    int32 status;
    int32 rank, nt, nattr;
    int32 dimSize[20];
    char  name[25];

    ENTER(2,"sdsGrokFile");
    /* make sure we don't crash on invalid open */
    ncopts = 0;

        
        /* Open temproary file to write HTML description of HDF/netCDF file */
        if (!(l_vars->hfp = fopen(tmp, "w")))
            return(NULL);

        /* Write MIME header */
        *mime_type = TEXT_HTML;
        if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL) {
          
            gateway_err(l_vars->hfp,"sdsGrokFile: writing HTML header",0);
            return NULL;
        }

        DBUG_PRINT(3,(LOGF,"Opening file = %s for Vxx \n",fname));
       
	/* set display mode */
	l_vars->display_mode = 2;
	
	/* Do SDSs */
	DBUG_PRINT(3,(LOGF,"sdsGrokFile: doing sds \n"));

	/* process SDS in file */
	if (do_sds(fname, l_vars) == FAIL) {
	  
	  gateway_err(l_vars->hfp,"sdsGrokFile: error processing SDS ",0);
	  return NULL;
	}

	/* get reference number */
	if (sscanf(l_vars->hdf_ref,"ref=%d,plane=%d",&ref,&l_vars->plane)==2) {
	  
	  DBUG_PRINT(3,(LOGF,"plane=%d \n", l_vars->plane));
	  
	  /* Open file for reading */
	  if ((sdid = SDstart(fname, DFACC_RDONLY)) == FAIL) {      
	  
	    DBUG_PRINT(3,(LOGF,"sdsGrokFile: Can not do SDstart\n"));
	    gateway_err(l_vars->hfp,"sdsGrokFile: opening file ",0);
	    return NULL;
	  }
	  
	  /* get index for the specified dataset by the ref. */
	  if (l_vars->do_netcdf) {     
            i = ref; /* for netCDF we use the index to indicate ref
                      * as they really don't have NDG */
          }
	  else {       
	    i = SDreftoindex(sdid, ref);
	    if (i == FAIL) {
	      DBUG_PRINT(3,(LOGF,"sdsGrokFile: Can not do SDreftoindex\n"));
	      return NULL;
	    }
	  }

	  /* select dataset */
	  sdsid = SDselect(sdid, i);
	  if (sdsid == FAIL) {
	    DBUG_PRINT(3,(LOGF,"sdsGrokFile: Can not do SDselect\n"));
	    gateway_err(l_vars->hfp,"do_sds: opening file ",0);
	    return NULL;
	  }
	  
	  
	  /* get dtatset info */
	  status = SDgetinfo(sdsid, name, &rank, dimSize, &nt, &nattr);
	  if (status == FAIL) {
	    DBUG_PRINT(3,(LOGF,"sdsGrokFile: SDgetinfo fail\n"));
	    return NULL;
	  }
	  else {
	    strcpy(l_vars->hdf_ref, " ");
	    strcat(l_vars->hdf_ref, name);
	  }
	  
	  
	  /* Terminate access to the array  dataset  */
	  SDendaccess(sdsid);
	  
	  /* Terminate access to the SD interface and close the file */
	  SDend(sdid);
	  
	}
	
	
	DBUG_PRINT(3,(LOGF,"Test l_vars->hdf_ref = %s  \n",l_vars->hdf_ref));
	
	  dump_sds(fname, current_entry, entries, num_entries,l_vars);
	
	  
	  /* Write MIME trailer */
	  *mime_type = TEXT_HTML;
	  if (write_html_trailer(l_vars->hfp, *mime_type,l_vars) == FAIL) {
	    
	    gateway_err(l_vars->hfp," writing HTML trailer",0);
	    return NULL;
	  }

    
    /* Close temporary file */
      fclose(l_vars->hfp);
    EXIT(2,"sdsGrokFile");
    return(tmp);
}

/*---------------------------------------------------------------------------
 NAME
       sdbGrokView
 DESCRIPTION
        The function sdbGrokView() will preview all of the planes within the image

        
 RETURNS
----------------------------------------------------------------------------*/
char *
sdbGrokView(char *fname,  char *ref,
            entry entries[MAX_ENTRIES], 
            mime_content_type *mime_type,
            int current_entry,
            int num_entries,
            lvar_st *l_vars)
{
    char *tmp = (char *)tempnam(NULL,"html");  /* Temporary file name for HTML desc */
    int32 count, start,end;            /* plane varies */
    int   i;
    int   refNumber;


    ENTER(2,"sdbGrokView");
     
    if (ref[0]==';')
        ref++;
  
    DBUG_PRINT(1,(LOGF,"ref = %s\n",ref )); 
   
    if(sscanf(ref, "ref=%d,start=%d,end=%d",&refNumber,&start,&end) == 3) {
	
      
            /* Open temproary file to write HTML description  */
            if (!(l_vars->hfp = fopen(tmp, "w")))
                return(NULL);

            /* Write MIME header */
            *mime_type = TEXT_HTML;
            if (write_html_header(l_vars->hfp, *mime_type,l_vars) == FAIL) {
          
                gateway_err(l_vars->hfp,"sdbGrokView: writing HTML header",0);
                return NULL;
            }

        /* read planes from file */
        /* Print Image header stuff in HTML  */
        count = end - start +1;

        print_label(l_vars->hfp, SDB1);
    
        fprintf(l_vars->hfp, "Total Number of Planes: %d(from:%d to:%d)<p>\n",
                count,start,end);

        /* For each plane in the file */
        for(i = start; i <= end; i++) {

            /* Print out image info in HTML depending on user prefrences */
            fprintf(l_vars->hfp,"<A HREF=\"%s%s?%s!sdbref;ref=%d,s=%d,plane=%d\"><IMG SRC=\"%s%s?%s!sdbref;ref=%d,s=%d,plane=%d\"></A> \n", 
                    l_vars->h_env->script_name,l_vars->h_env->path_info,l_vars->f_name,
                    refNumber, 0,i, l_vars->h_env->script_name, 
                    l_vars->h_env->path_info,l_vars->f_name, refNumber, 1, i);
        }
    } 
    else {
        return NULL;
    }
   
    
    /* Close temporary file */
        fclose(l_vars->hfp);
  
    EXIT(2,"sdbGrokView");
    return(tmp);
    
}  

