/****************************************************************************
 * NCSA HDF                                                                 *
 * Software Development Group                                               *
 * National Center for Supercomputing Applications                          *
 * University of Illinois at Urbana-Champaign                               *
 * 605 E. Springfield, Champaign IL 61820                                   *
 *                                                                          *
 * For conditions of distribution and use, see the accompanying             *
 * COPYING file.                                                            *
 *                                                                          *
 ****************************************************************************/

#ifndef SDB_H
#define SDB_H

/* --------------- includes  ----------------------------- */
#if defined(hpux) || defined(__hpux)
#include <sys/syscall.h>
#define getrusage(a, b)  syscall(SYS_GETRUSAGE, a, b)
#endif /* hpux */

/* required headers */
#include <stdio.h>
#ifdef _POSIX_SOURCE
#include <sys/stat.h>
#endif /* _POSIX_SOURCE */

/* To use utility macros...*/

#include "myutil.h" 
#include "fbm.h"          /* For gif handling code */
#include "mfhdf.h"
#include "hdf.h"
#include "herr.h"         /* To get some decent error reporting...*/
#include "vg.h"
#include "glist.h"
extern char *strdup(const char *);

/*---------------------- Local defines -----------------------
 * The following info from should be sent from the browser 
 * But until forms are used they are put here
 */

/* make all in-lined images smaller than hdfImageSize X hdfImageSize */
/* #define hdfImageSize Rdata.hdf_max_image_dimension */
#define hdfImageSize 200

/* limits on datasets and attribute counts to send into brief mode */
/* #define MAX_DATASET_DISPLAY   Rdata.hdf_max_displayed_datasets */
#define MAX_DATASET_DISPLAY   10

/* #define MAX_ATTRIBUTE_DISPLAY Rdata.hdf_max_displayed_attributes */
#define MAX_ATTRIBUTE_DISPLAY 10


/* Send reply buffer size */
#define SND_BUF_SIZE 524288

/* hmm...eric what limits are these? */
#define NUM_VGS       20
#define DOC_WIDTH   132  /* Width of the html document for display a tree */
#define MAX_ENTRIES 10000
#define strequal(s1,s2)    (!strcmp((s1),(s2)))


#define SDBREF    1
#define SDBPLANE  2
#define SDBVIEW   3
#define SDSVIEW   31    /* for hdf sds */
#define LISTHEAD  10
#define HEAD      11
#define HISTORY   12

#define V_IMAGE 1
#define V_ASC_FILE 2
#define V_HDF_FILE 3
#define V_BINARY_FILE 4
#define V_TXT 5

/*-------------------- Local structures -----------------------------*/

/* Stole this structure framework for handling Images from libhtml/HTML.h */
typedef struct image_rec {
    int ismap;                 /* Is this an ISMAP */
    int width;                 
    int height;
    int num_colors;            /* 256 we handle only raster8 */
    int *reds;                 /* R */
    int *greens;               /* G */
    int *blues;                /* B */
    unsigned char *image_data; /* actual data */
    unsigned char *color_map;  /* palette */
} ImageInfo, *ImageInfoPtr;

/* from util.c - used in main() */
typedef struct {
    char *name;
    char *val;
} entry;

/* Enumerated types for MIME */
typedef enum {
  TEXT_HTML,       /* plain html text */
  IMAGE_GIF,       /* a gif image */
  APP_HDF,         /* application HDF file, inserted by lpd */
  APP_BIN,         /* application binary file--lpd */
  APP_TXT          /* plain text file --lpd */
} mime_content_type;

/* This structure is used to get all relavant environment variables
   passed to this CGI program by 'httpd' */
typedef struct hserv_t {
  char *request_method;      /* GET / PUT */
  char *path_info;
  char *path_translated;
  char *script_name;         /* CGI script name */
  char *query_string;        /* used by GET */
  char *remote_host;
  char *remote_addr;
  char *auth_type;
  char *remote_user;
  char *content_type;
  char *content_length;
  char *http_accept;
  char *http_user_agent;
} hserv_st, *hserv_stp;


/* Create a local struct to hold local variables that can be passed around */
typedef struct lstruct_t {
    FILE *hfp;      /* File pointer to temproray HTML description */
    FILE *gfp;      /* File pointer to temproray gif image */
    int    brief;    /* brief flag */
    char * my_url;   /* Pointer to URL ? */

    /* Environment */
    hserv_st *h_env;  /* Structure holding server environment variables */

    /* Path related */
    char *hdf_path;   /* hdf path w/ filename to show in URL? */
    char *hdf_path_t; /* tranlsated relative path to file passed to script by httpd*/
    char *f_name;     /* name of HDF/netCDF file being browsed */
    char *f_path_r;   /* relative path to HDF/netCDF file being browsed */
    char *hdf_file;   /* full HDF file name w/path if specified */
    char *hdf_file_r; /* name of HDF/netCDF w/ relative path */

    /* Object stuff */
    char *hdf_ref;    /* current 'hdf_ref' i.e.tag/ref in HDF/netCDF file */
    uint16 obj_tag;   /* object tag from 'hdf_ref' */
    uint16 obj_ref;   /* object ref from 'hdf_ref' */
    int    sub_s;     /* subsetting/sampling flag from 'hdf_ref' */


    /* Used to indicate which way to display data, in scroll boxes, */
    /* in a form with fields, display raw data */
    int display_mode;       /* Values 1,2 or 3 */

    /* SDS - subsetting */
    int32 *sd_dim_start;    /* array for starting dimensions */
    int32 *sd_dim_stride;   /* array for strides */
    int32 *sd_dim_end;      /* array for ending dimensions */
    int32 ndims;            /* numer of dimesions read in */

    /* Vdata field name selecting, subsetting, and plotting */
    int *field_indices;     /* array of selected Vdata field names(indices)
                              if an indices value is 0, all fields are selected--lpd */
    int nfields;            /* nubmer of valid fields */
    int ntfields;           /* number of total fields in the vdata--lpd*/
    char *xyplot_name;      /* the plot title */
    char *x_axis_name;      /* the x-axis name */
    char *y_axis_name;      /* the y-axis name */
    int *yfield_indices;    /* array of selected Vdata fields for Y axis */
    int nyfields;           /* number of y fields selected */
    int vh_start_rec;       /* starting record for subsetting */
    int vh_end_rec;         /* ending record for subsetting */
    int vh_skip_rec;        /* skipping record for subsetting --lpd*/
    int scaling;            /* scaling methods for both xyplot and sds image */
    int xyplot_style;       /* palette for SDS, plot line style for xy plot */
    int view_type;          /* view type for the subsetting result, for both SDS and Vdata --lpd */

    int  plane;            /* plane number to be viewed */

    /* FITS PHDU or IMAGE extension variables regarding the subsetting  
       the following variables are  the common variable shared by fits
       and sds(convert to image */
    int sdb_ref;            /* the reference number  */
    int sdb_imgFlag;        /*  do   image subsetting   */   
    int sdb_plane;          /* nubmer of planes    */
    int sdb_start_rec;      /* starting record for subsetting */
    int sdb_end_rec;        /* ending record for subsetting       */

    /* netCDF */
    int  do_netcdf;        /* netcdf flag */

    /* SDS lists */
    Generic_list  *sds_list;    /* List of both dimension and regualr variables */
    Generic_list  *gattr_list;  /* global attribute List */
    Generic_list  *dimvar_list; /* dimension variable List */
    Generic_list  *regvar_list; /* regular variable List */

    /* palette list */
    Generic_list  *pal_list;  /* Palette List */

    /* raster list */
    Generic_list  *rig_list;  /* Raster8 List */

    /* annotation lists */
    Generic_list  *an_list;   /* all annotations List */
    Generic_list  *flan_list; /* file label List */
    Generic_list  *fdan_list; /* file desc List */
    Generic_list  *olan_list; /* object label List */
    Generic_list  *odan_list; /* object desc List */

    /* Vgroup list */
    Generic_list  *vg_list;   /* all NON-SDS Vgroups */
    Generic_list  *tvg_list;  /* Top level Vgroup list */
    Generic_list  *ovg_list;  /* Not top level Vgroup */

    /* Vdata list */
    Generic_list  *vd_list;  /* Vdata List */

} lvar_st, *lvar_stp;

/* prototype */
IMPORT char * sdbGrokFile(char *fname,    /* HDF file name */
                               entry entries[MAX_ENTRIES], 
                               mime_content_type *mime_type,
                               int current_entry,
                               int num_entries,
                               lvar_st *l_vars);
IMPORT char *sdbGrokRef(char *fname,                 /* HDF/netCDF file name */
                        char *ref,                  /* Reference number to start navigation from */
                        entry entries[MAX_ENTRIES], 
                        mime_content_type *mime_type, 
                        int current_entry, 
                        int num_entries,
                        lvar_st *l_vars);
IMPORT char *sdbGrokView(char *fname,  char *ref,
                         entry entries[MAX_ENTRIES], 
                         mime_content_type *mime_type,
                         int current_entry,
                         int num_entries,
                         lvar_st *l_vars);
IMPORT char *sdsGrokRef(char *fname, 
           char *ref,
           entry entries[MAX_ENTRIES], 
           mime_content_type *mime_type, 
           int current_entry, 
           int num_entries,
           lvar_st *l_vars);
IMPORT char *sdsGrokFile(char *fname, 
            entry entries[MAX_ENTRIES], 
            mime_content_type *mime_type,
            int current_entry,
            int num_entries,
            lvar_st *l_vars);

#endif /* SDB_H */
