
/*--------------------------------------------------------------------
 NAME
       sds_subset
 DESCRIPTION
       subseting the SDS's
 RETURNS
       file name for the subseted data.
 AUTHOR
       Dr. Liping Di, Hughes STX.  
 COMMENTS
 
 Copyright (C) 1997 Hughes STX Corporation.
 Permission to use, modify, and distribute this software and its 
 documentation for any purpose without fee is hereby granted, provided 
 that the above copyright notice appear in all copies and that both that 
 copyright notice and this permission notice appear in supporting 
 documentation.

---------------------------------------------------------------------*/
#define V_IMAGE 1
#define V_ASC_FILE 2
#define V_HDF_FILE 3
#define V_BINARY_FILE 4
#define V_TXT 5
#define S_NONE 1
#define S_LINEAR 2
#define S_HISTO 3

#include "sdb.h"
#include "sdb_sds.h"
#include "sdb_attr.h"
#include "sdb_util.h"
#include "html_util.h"

int data_scaling(void *buffer, int32 nt, int n_elements, int n_levels, int scaling);
void histo_eq(void *buffer, int n_elements, int type, long histo[], int n_bin, int n_levels);
int GetGIFPixel_my(int, int);
unsigned char *pixel_buffer;
int mywidth; 

char *
sds_subset(char *fname, int current_entry, entry entries[MAX_ENTRIES], 
       int num_entries,lvar_st *l_vars, mime_content_type *mime_type)
{
   int32 sdid;     /* file handle */
   int32 sdsid;    /* Data set handle */
   int32 index;    /* SDS index */
   int32 i, j;  /* loop variables */
   char name[512]; /* sds name */
   int32 nt;             /* Number type */
   int32 dimsizes[MAXDIM];   /* array for dimension sizes, fixed at 50 */ 
   int32 subdim[MAXDIM], substart[MAXDIM];     /* for the subset dimension */
   int32 rank;           /* rank of data set */
   int32  count;         /* number of attributes */
   char *ftmp;           /* temporary file for file output */
   int display_mode;
   int32 status, ndattr, rank_tmp, t_nt, acount;
   int32 sdtmp=FAIL;
   int32 sdstmp=FAIL, no_stride;
   size_t size_bytes, bytes, bytes_step;
   unsigned char *ptr1, *ptr2;
   float step;
   unsigned char value;
   char *name_only;
   
 
   FILE *h_fp, *f_fp;
   char tmp_html[32768];
   char t_name[MAX_NC_NAME];
   int view_type, scaling, n_levels;

   
   ENTER(2,"sds_subset");
   DBUG_PRINT(1,(LOGF," l_vars->display_mode=%d\n",l_vars->display_mode)); 
   DBUG_PRINT(3,(LOGF," before l_vars->hdf_ref=%s\n",l_vars->hdf_ref)); 

   /* get the display_mode */
   display_mode = l_vars->display_mode;
   if(display_mode != 3) return NULL;               /* not get the real data */
      
   DBUG_PRINT(1,(LOGF," fname=%s\n",fname)); 
   DBUG_PRINT(1,(LOGF," display_mode=%d\n",display_mode)); 
   DBUG_PRINT(3,(LOGF," after l_vars->hdf_ref=%s\n",l_vars->hdf_ref)); 
   
   name_only=strrchr(fname, '/');
   if(name_only)
      name_only ++;
   else
      name_only = fname;
      
   /* check the view type */
   /* for(i=0; i<num_entries; i++){
      if( strcmp(entries[i].name, "retrieve type") == 0){
             view_type = atoi(entries[i].val);
             break;
      }
   }
   */ 

  view_type = l_vars->view_type;
  if(l_vars->view_type == V_TXT)return NULL; /* no view type or the html text type */
  
   h_fp = l_vars->hfp;
   /* Open file for reading */
   if ((sdid = SDstart(fname, DFACC_RDONLY)) == FAIL){
      sprintf(tmp_html, "sds_subset:can not open HDF file %s", fname);
      gateway_err(h_fp,tmp_html,1);
   }

   if((index = SDreftoindex(sdid, l_vars->obj_ref)) == FAIL){
      sprintf(tmp_html, "sds_subset: Can not find SDS %s", l_vars->hdf_ref);
      gateway_err(h_fp, tmp_html, 1);
   }
   
   /* select the SDS id */
   if((sdsid = SDselect(sdid, index)) == FAIL){
      sprintf(tmp_html, "sds_subset: can not select SDS %s", l_vars->hdf_ref);
      gateway_err(h_fp, tmp_html, 1);
   }
   
   if((status = SDgetinfo(sdsid, name, &rank, dimsizes, &nt, &ndattr)) == FAIL){
      SDendaccess(sdsid);
      SDend(sdid);
      gateway_err(h_fp,"sds_subset: can not getting dataset info \n",1);
   }

   if (rank>0){
        int32 *start, *stride, *edge, *left;
        int ret, r, end;
        size_t size=1;
        void *buffer;
        int32 width, height, m1st, m2nd;
        int reds[256], greens[256], blues[256];
         
        start  = (int32 *)HDgetspace(rank * sizeof(int32));
        stride = (int32 *)HDgetspace(rank * sizeof(int32));
        edge   = (int32 *)HDgetspace(rank * sizeof(int32));
        left   = (int32 *)HDgetspace(rank * sizeof(int32));
        if (start == NULL || stride == NULL || edge == NULL || left == NULL){
            SDendaccess(sdsid);
            SDend(sdid);
            gateway_err(h_fp,"do_sds: failed to allocate space \n",1);
        }
          else {
            /* Get subsampling values along each dimension */
            for (r=0; r<rank; r++){
                 DBUG_PRINT(1,(LOGF, "dump_sds: sd_dim_start[%d]=%d \n",
                          r,l_vars->sd_dim_start[r]));
                 DBUG_PRINT(1,(LOGF, "dump_sds: sd_dim_stride[%d]=%d \n",
                          r,l_vars->sd_dim_stride[r]));
                 DBUG_PRINT(1,(LOGF, "dump_sds: sd_dim_end[%d]=%d \n",
                          r,l_vars->sd_dim_end[r]));
                
                start[r] = l_vars->sd_dim_start[r];
                if ((stride[r] = l_vars->sd_dim_stride[r])==0)
                    stride[r] = 1;
                if ((end = l_vars->sd_dim_end[r])==0)
                    left[r] = dimsizes[r] - start[r];
                else
                    left[r] = end - start[r];
                
                edge[r] = 1;
            }
            
            edge[rank-1] = left[rank-1];
               switch (view_type){
               case V_BINARY_FILE:
                   for(i=0; i<rank; i++){
                         edge[i]= left[i]/stride[i];
                         if(left[i]%stride[i] != 0)edge[i]=edge[i]+1;
                         if(edge[i] == 0)edge[i]=1;
                         size = size * edge[i];
                   }
                   size_bytes = size *DFKNTsize(nt);
                   size = (size_bytes/sizeof(double)+1)*sizeof(double);
                   DBUG_PRINT(1, (LOGF, "\nsize=%d\n", size));
                   buffer = (void *)HDgetspace(size);
                   if(buffer == NULL){
                       SDendaccess(sdsid);
                       SDend(sdid);
                       gateway_err(h_fp,"sds_subset: failed to allocate temporary memory\n",1);
                    }
               
                    /* open the temporary binary file */
                    ftmp = (char *)tempnam(NULL, "bin");
                    if((f_fp = fopen(ftmp, "wb+")) == NULL){
                        free(ftmp);
                        HDfreespace(buffer);
                        SDendaccess(sdsid);
                        SDend(sdid);
                        gateway_err(h_fp, "sds_subset: failed to create temporary binary file", 1);
                    }
                
                    SDreaddata(sdsid, start, stride, edge, buffer);
                    fwrite(buffer, 1, size_bytes, f_fp);
                    HDfreespace(buffer);
                    fclose(f_fp);
                    *mime_type = APP_BIN;
                    printf("Content-type: application/octet-stream;name=\"test.bin\"%c",10);
                    printf("File-name: test.bin%c%c",10,10);
#ifdef WIN32
		            setoutput_binarymode();
#endif /* WIN32 */
                    break;
                                                         
               case V_HDF_FILE:
                    rank_tmp=0;
                    for(i=0; i<rank; i++){
                         edge[i]=left[i]/stride[i];
                         if(left[i]%stride[i] !=0)edge[i]=edge[i]+1;
                         if(edge[i]==0)edge[i]=1;
                         if(edge[i] > 1){
                              substart[rank_tmp]=0;
                              subdim[rank_tmp]=edge[i];
                              rank_tmp++;
                         }
                         size=size*edge[i];
                    }
                    if(rank_tmp == 0)rank_tmp=1;
                    size = (size*DFKNTsize(nt)/sizeof(double)+1)*sizeof(double);
                    DBUG_PRINT(1, (LOGF, "\nsize=%d\n", size));
                    buffer = (void *)HDgetspace(size);
                    if(buffer == NULL){
                        SDendaccess(sdsid);
                        SDend(sdid);
                        gateway_err(h_fp,"sds_subset: failed to allocate temporary memory\n",1);
                    }
               
                    /* open the temporary HDF file */
                    ftmp = (char *)tempnam(NULL, "hdf");
                
                    if((sdtmp=SDstart(ftmp, DFACC_CREATE))==FAIL){
                        free(ftmp);
                        HDfreespace(buffer);
                        SDendaccess(sdsid);
                        SDend(sdid);                     
                        gateway_err(h_fp, "sds_subset: failed to create temporary HDF file", 1);
                    }
                
                    /* get the data */
                    sprintf(tmp_html, "This HDF file, created by DAAC-in-a-Box,\n"
                                  "is a subset of original HDF file: %s.\n"
                                  "All other global attributes are directly copied from\n"
                                  "the original file.", name_only); 
                    SDsetattr(sdtmp, "DIB data source", DFNT_CHAR8, strlen(tmp_html), tmp_html);

                    /* copy all global attributes */
                    SDfileinfo(sdid, &t_nt, &count);
                    for(j = 0; j < count; j++){
                          SDattrinfo(sdid, j, t_name, &t_nt, &acount);
                          SDreadattr(sdid, j, tmp_html);
                          SDsetattr(sdtmp, t_name, t_nt, acount, tmp_html);
                    }

                    if((sdstmp=SDcreate(sdtmp, name, nt, rank_tmp, subdim)) == FAIL){
                        HDfreespace(buffer);
                        SDend(sdtmp);
                        remove(ftmp);
                        free(ftmp);
                        SDendaccess(sdsid);
                        SDend(sdid); 
                        gateway_err(h_fp, "sds_subset: failed to create subseted SDS", 1);
                    }
                
                    /* write the subseted data */               
                    if (SDreaddata(sdsid, start, stride, edge, buffer) == FAIL){
                        HDfreespace(buffer);
                        SDend(sdtmp);
                        remove(ftmp);
                        free(ftmp);
                        SDendaccess(sdsid);
                        SDend(sdid); 
                        gateway_err(h_fp, "sds_subset: failed to read the subseted SDS", 1);
                    }
                    SDwritedata(sdstmp, substart, NULL, subdim, buffer);
                
                    sprintf(tmp_html, "The SDS \"%s\" is a subset of the same SDS\n"
                            "in the original HDF file: %s.\n" 
                            "Except for the subset start, stride, and edge, all other\n"
                            "attributes are directly copied from the original local attributes", name, name_only);  
                    SDsetattr(sdstmp, "Subsetting description", DFNT_CHAR8, strlen(tmp_html), tmp_html);
                    SDsetattr(sdstmp, "subset start", DFNT_INT32, rank, start);
                    SDsetattr(sdstmp, "subset stride", DFNT_INT32, rank, stride);
                    SDsetattr(sdstmp, "subset edge", DFNT_INT32, rank, edge);
                
                    /* copy all local attribute */
                    for(j = 0; j < ndattr; j++){
                         SDattrinfo(sdsid, j, t_name, &t_nt, &acount);
                         SDreadattr(sdsid, j, tmp_html);
                         SDsetattr(sdstmp, t_name, t_nt, acount, tmp_html);
                    }
                    HDfreespace(buffer);
                    SDendaccess(sdstmp);
                    SDend(sdtmp);
                    fclose(h_fp);
                    *mime_type = APP_HDF;
                    printf("Content-type: application/x-hdf%c",10);
                    printf("Name: test.hdf%c%c",10,10);
#ifdef WIN32
		            setoutput_binarymode();
#endif /* WIN32 */
                    break;                    

               case V_IMAGE:
                    
                    no_stride=1;
                    for (i=0; i<rank; i++){
                        edge[i]=left[i]/stride[i];
                        if(left[i]%stride[i] != 0)edge[i]=edge[i]+1;
                        if(edge[i]==0)edge[i]=1;
                        if(stride[i]>1)no_stride=0;
                    }
                   
                    /* get the two display dimension */
                    if(rank == 1){
                        width = edge[0];
                        height= 1;
                    }
                    else {
                        width = edge[1];
                        height = edge[0];
                        m1st=0;
                        m2nd=1;
                        if(rank > 2){
                             m1st = 0;
                             for(i=1; i<rank; i++){
                                  if(height <= edge[i]){
                                       m1st = i;
                                       height=edge[i];
                                  }
                             }
                             height= -10;
                             m2nd = 0;
                             for(i=0; i< rank; i++){
                                  if(i != m1st && height <= edge[i]){
                                       m2nd = i;
                                       height = edge[i];
                                  }
                             }
                             if(m1st > m2nd){
                                  width = edge[m1st];
                                  height = edge[m2nd];
                                  i=m2nd;
                                  m2nd=m1st;
                                  m1st=i;
                             }
                             else {
                                  width = edge[m2nd];
                                  height = edge[m1st];
                             }
                             for(i=0; i < rank; i++)
                                  if(i != m1st && i != m2nd)edge[i]=1;
                        }
                    }               
                
                    /* allocate the memory */
                    bytes=DFKNTsize(nt);
                    if(no_stride)
                        size = width * height * bytes;
                    else
                        size =(width * height + dimsizes[m2nd])*bytes;
                    if(size < width * (height + 20))size = width *( height + 20);

                    size = (size/sizeof(double)+1)*sizeof(double);

                    sprintf(tmp_html, "\nsize=%d\n", size);
                    DBUG_PRINT(1, (LOGF, tmp_html));
                    buffer = (void *)HDgetspace(size);
                    if(buffer == NULL){
                        SDendaccess(sdsid);
                        SDend(sdid);                     
                        gateway_err(h_fp,"sds_subset: failed to allocate temporary memory\n",1);
                    }
                    
                    sprintf(tmp_html, "\nStart=%d, %d, %d, %d, %d, %d\n", start[0], start[1], stride[0], stride[1], edge[0],edge[1]);
                    DBUG_PRINT(1,(LOGF,tmp_html));
                    if(no_stride){
                         if( SDreaddata(sdsid, start, NULL, edge, buffer) == FAIL){
                             DBUG_PRINT(1,(LOGF,"\nFail to get data"));
                             HDfreespace(buffer);
                             SDendaccess(sdsid);
                             SDend(sdid);
                             gateway_err(h_fp,"sds_subset: failed to read the data\n",1);
                         }
                    }
                    else {
                         ptr1 = ptr2 = buffer;
                         edge[m1st]=1;
                         bytes_step=stride[m2nd]*bytes;
                         DBUG_PRINT(1, (LOGF, "bytes=%d, bytes_step=%d, m1st=%d, m2nd=%d\n", bytes, bytes_step, m1st, m2nd));
                         edge[m2nd]=dimsizes[m2nd]-start[m2nd];
                         for(i=0; i<height; i++){
/*                           DBUG_PRINT(1, (LOGF, "start=%d, %d, %d, %d\n", start[0],start[1], edge[0], edge[1]));
*/
                             if( SDreaddata(sdsid, start, NULL, edge, ptr2) == FAIL){
                                  DBUG_PRINT(1,(LOGF,"\nget data failed"));
                                  HDfreespace(buffer);
                                  SDendaccess(sdsid);
                                  SDend(sdid);
                                  gateway_err(h_fp,"sds_subset: failed to read the data\n",1);
                              }
                              else {
                                   for(j=0; j<width; j++){
                                        memcpy(ptr1, ptr2, bytes);
                                        ptr1 += bytes;
                                        ptr2 += bytes_step;
                                   }
                                   ptr2=ptr1;
                                   start[m1st] += stride[m1st];
                              }
                         }
                         if(i != height) break;          
                    }

                    /* find palette */
                    *tmp_html='\0';
                    for(i=current_entry; i< num_entries; i++){
                         if(strcmp(entries[i].name, "palette") == 0){
                              strcpy(tmp_html, entries[i].val);
                              break;
                         }
                    }
                    
                    /* get the scaling method */
                    scaling = l_vars->scaling;
/*                    for(i=current_entry; i< num_entries; i++){
                         if(strcmp(entries[i].name, "scaling") == 0){
                              scaling = atoi(entries[i].val);
                              break;
                         }
                    }          
*/                    
                    /* fill in the palette, initial as bw */ 
                    for(i = 0; i< 256; i++){
                        reds[i] = i;
                        greens[i] = i;
                        blues[i] = i;
                    }

                    /* fill in the palette */
                    n_levels=255;
                    if(*tmp_html != '\0'){
                         if(strstr(tmp_html, "HDF_Internal_")){   /* internal palette */
                             int which_pal = atoi(strrchr(tmp_html, '_')+1);
                             if(DFPrestart() == FAIL)
                                  DBUG_PRINT(1, (LOGF, "reset the palette failed"));
                             for(i=0; i<which_pal; i++){
                                  if (DFPgetpal(fname, tmp_html) == FAIL)
                                       DBUG_PRINT(1, (LOGF, "read the %dth pal failed", i));
                                  else
                                       DBUG_PRINT(1, (LOGF, "read successful"));
                             }
                             ptr1=(unsigned char *)tmp_html;
                             DBUG_PRINT(1,(LOGF,"which_pals =%d, file=%s\n", which_pal, fname));
                             for(i=0; i<256; i++){                             
                                  reds[i] = *(ptr1 + 3*i);
                                  greens[i]= *(ptr1 + 3*i+1);
                                  blues[i]= *(ptr1 + 3*i+2);
                                  DBUG_PRINT(1, (LOGF, "%d %d %d %d\n", i, reds[i], greens[i], blues[i]));
                             }
                         } 
                         else if (strcmp(tmp_html, "Black-White") != 0){  /* from server palette */
                              int r, g, b;
                              FILE *pal;
                              strcat(tmp_html, ".pal");
                              pal = fopen(tmp_html, "r");
                              n_levels=0;
                              if(pal){
                                   while (fscanf(pal,"%d %d %d %d ", &i, &r, &g, &b) == 4){
                                        if(i >=0 && i < 256){
                                             reds[i]=r;
                                             greens[i]=g;
                                             blues[i]=b;
                                            if(i > n_levels)n_levels=i;
                                        }
                                   }
                                   fclose(pal);     
                              }
                         }
                    }
               
                    /* handling the scaling methods */                
                    if(data_scaling(buffer, nt, width*height, n_levels, scaling)== -1){
                         HDfreespace(buffer);
                         gateway_err(h_fp,"sds_subset: unrecognized data type\n", 0);
                         break;
                    } 

                    /* write the color bar */
                    step = (float)(n_levels+1)/width;
                    ptr1 = (unsigned char *)buffer + width * height;
                    memset(ptr1, 0, 10 * width);
                    ptr1 = ptr1 + 10 * width;
                    for(i=0; i<width; i++){
                         value = step * i;
                         for(j=0; j< 10; j++)
                              *(ptr1+j*width)= value;
                         ptr1++;
                    }
                    height += 20;
                                              
                    /* open the temporary file for the gif file */
                    ftmp = (char *)tempnam(NULL, "gif"); 
                    f_fp=l_vars->gfp = fopen(ftmp, "wb+");
                    if(f_fp == NULL ){
                         HDfreespace(buffer);
                         SDendaccess(sdsid);
                         free(ftmp);
                         SDend(sdid);
                         gateway_err(h_fp,"sds_subset: failed to create temporary gif file\n", 1);
                         break;
                    }
                
                    /* gif encoding */
                    /* mypixels is a gloabl variable which is the address for the image buffer */
                    /* it is used by myGetGIFPixel to get the next values, so it have to be set to */
                    /* buffer before calling to GIFEncode. The same for myrowlen.  bad coding method */
                    pixel_buffer = buffer;
                    mywidth = width;
                    GIFEncode(l_vars->gfp, width, height, 0, 0, 8, reds, greens, blues, GetGIFPixel_my);
                    HDfreespace(buffer);
                
                    /* set out the header info */
                    *mime_type = IMAGE_GIF;
                    fclose(f_fp);
                    printf("Content-type: image/gif%c%c",10,10);
                    break;
               
               case V_ASC_FILE:
                    ftmp = (char *)tempnam(NULL, "ASC");
                    if((f_fp = fopen(ftmp, "w+")) == NULL){
                        free(ftmp);
                        SDendaccess(sdsid);
                        SDend(sdid);
                        gateway_err(h_fp, "sds_subset: failed to create temporary ASCII file", 1);
                    }
                
                    ret = sdsdumpfull(sdsid, rank, dimsizes, nt, 0, f_fp, 
                             start, stride, edge, left, l_vars); 
                
                    fclose(f_fp);
                    if (ret == FAIL){
                         SDendaccess(sdsid);
                         SDend(sdid);
                         remove(ftmp);
                         free(ftmp);
                         gateway_err(h_fp,"do_sds: failed to dump sds into an ASCII file \n",1);     
                    }
                    else {
                         *mime_type = APP_TXT;
                         printf("Content-type: application/plain-text;name=\"test.ASC\"%c",10);
                         printf("File-name: test.ASC%c%c",10,10);
                    }
                    break;
                
            default:
                    SDendaccess(sdsid);
                    SDend(sdid);
                    gateway_err(h_fp,"do_sds: unknown view type \n",1);
                    break;
            }

            /* free up space */
            HDfreespace(start);
            HDfreespace(stride);
            HDfreespace(edge);
            HDfreespace(left);
        } /* display_mode=3 */
   }
    
   SDendaccess(sdsid);
   SDend(sdid);
   return ftmp;
   EXIT(2,"sds_subset");
} /* do_sds */

int
GetGIFPixel_my (int x, int y)
{
  return (pixel_buffer[y * mywidth + x]);
} /* myGetGIFPixel */

#define N_BIN 512
int
data_scaling(void *buffer, int32 nt, int n_elements, int n_levels, int scaling) 
{
unsigned char look_up[256];
unsigned char *pbuf, u8max, u8min;
long histo[N_BIN];

char *ptr_char, c8max, c8min;
int16 *ptr_i16, i16max, i16min;
uint16 *ptr_u16, u16max, u16min;
int32 *ptr_i32, i32max, i32min;
uint32 *ptr_u32, u32max, u32min;
float32 *ptr_f32, f32max, f32min, a1, b1;
float64 *ptr_f64, f64max, f64min;
float64 a, b;
int16 *p2buf;

int i;
  
  pbuf = (unsigned char *)buffer;
  p2buf = (int16 *)buffer;
  if(scaling == S_HISTO){
       for(i=0; i<N_BIN; i++)
            histo[i]=0;
  }
  
  switch(nt){
  
  case DFNT_CHAR8:
  case DFNT_INT8:
       
            ptr_char = (char *)buffer;
            
            switch(scaling){
            
            case S_NONE:
                 for(i=0; i< n_elements; i++){
                      *pbuf =(unsigned char)((short)(*ptr_char) + 128);
                      pbuf ++;
                      ptr_char ++;
                 }
                 break;
            
            case S_HISTO:
                 for(; ptr_char < (char *)buffer + n_elements; ptr_char++){
                    *pbuf = (unsigned char)(*ptr_char + 128);
                    histo[*pbuf] ++ ;
                    pbuf ++ ;
               }
               histo_eq(buffer, n_elements, 1, histo, 256, n_levels);
               break; 

            case S_LINEAR:

                 c8max = c8min = *ptr_char;
                 for(; ptr_char < (char *)buffer + n_elements; ptr_char++){
                      if(*ptr_char > c8max)c8max = *ptr_char;
                      if(*ptr_char < c8min)c8min = *ptr_char;
                 }
                 if(c8max == c8min)break;
                 
                 a =(float32)n_levels/(c8max-c8min);
                 b = -a*c8min;
                 
                 for(i=c8min; i<= c8max; i++)
                      look_up[i-c8min] = (unsigned char )(a*i+b+0.1);
                 
                 ptr_char = (char *)buffer;
                 for(i = 0; i< n_elements; i++){
                      *pbuf = look_up[*ptr_char - c8min];
                      pbuf ++;
                      ptr_char ++;
                 }
                 
                 break;
            
            default:
                 break;
            }           
            
            break;     

  case DFNT_UCHAR8:
  case DFNT_UINT8:
            switch(scaling){
            
            case S_NONE:
               break;
          
            case S_HISTO:
               for(; pbuf < (unsigned char *)buffer + n_elements; pbuf++)
                    histo[*pbuf] ++;
                    
               histo_eq(buffer, n_elements, 1, histo, 256, n_levels);
               break; 
                 
            case S_LINEAR:
                 u8max = u8min = *pbuf;
                 for(; pbuf < (unsigned char *)buffer + n_elements; pbuf++){
                      if(*pbuf > u8max)u8max = *pbuf;
                      if(*pbuf < u8min)u8min = *pbuf;
                 }
                 if(u8max == u8min)break;

                 a =(float64)n_levels/(u8max - u8min);
                 b = -a*u8min;
                 
                 for(i=u8min; i<= u8max; i++)
                      look_up[i-u8min] = (unsigned char )(a*i+b+0.1);
                 
                 pbuf = (unsigned char *)buffer;
                 for(i = 0; i< n_elements; i++){
                      *pbuf = look_up[*pbuf - u8min];
                      pbuf ++;
                 }
                 break;
            default:
                 break;
            }           
            break;     

  case DFNT_UINT16:
  
            ptr_u16 = (uint16 *)buffer;
            u16max = u16min = *ptr_u16;
            for(; ptr_u16 < (uint16 *)buffer + n_elements; ptr_u16++){
                 if(*ptr_u16 > u16max)u16max = *ptr_u16;
                 if(*ptr_u16 < u16min)u16min = *ptr_u16;
            }
            if(u16max == u16min){
                 for(i=0; i< n_elements; i++){
                      *pbuf=n_levels;
                      pbuf++;
                 }
                 break;
            }                           
                 
            if(scaling == S_HISTO)
                 a1 = (float32)(N_BIN-1)/(u16max - u16min);
            else
                 a1 = (float32)n_levels/(u16max - u16min);
            b1 = (-a1*u16min) + 0.5;
            
            ptr_u16 = (uint16 *)buffer;                 
            if(scaling == S_HISTO){
                 for(i=0; i< n_elements; i++){
                      *p2buf = (int16)(*ptr_u16*a1 + b1);
                      histo[*p2buf] ++;
                      p2buf ++;
                      ptr_u16++;
                 }
                 histo_eq(buffer, n_elements, 2, histo, N_BIN, n_levels);
            }
            else {
                 for(i = 0; i< n_elements; i++){
                      *pbuf = (unsigned char)(*ptr_u16*a1 + b1);
                      pbuf ++;
                      ptr_u16++;
                 }
            }
            
            break;     
                 
  case DFNT_INT16:
  
            ptr_i16 = (int16 *)buffer;
            i16max = i16min = *ptr_i16;
            for(; ptr_i16 < (int16 *)buffer + n_elements; ptr_i16++){
                 if(*ptr_i16 > i16max)i16max = *ptr_i16;
                 if(*ptr_i16 < i16min)i16min = *ptr_i16;
            }
            if(i16max == i16min){
                 for(i=0; i< n_elements; i++){
                      *pbuf = n_levels;
                      *pbuf ++;
                 }
                 break;
            }
                 
            if(scaling == S_HISTO)
                 a1 = (float32)(N_BIN-1)/(i16max - i16min);
            else
                 a1 =(float32)n_levels/(i16max - i16min);
            b1 = (-a1*i16min) + 0.5;

            ptr_i16 = (int16 *)buffer;
            
            if(scaling == S_HISTO){
                 for(i=0; i<n_elements; i++){
                      *p2buf = (int16) (*ptr_i16 * a1 + b1);
                      histo[*p2buf] ++;
                      p2buf++;
                      ptr_i16 ++;
                 }
                 histo_eq(buffer, n_elements, 2, histo, N_BIN, n_levels);
            }
            else {
                 for(i = 0; i< n_elements; i++){
                      *pbuf = (unsigned char)(*ptr_i16*a1 + b1);
                      pbuf ++;
                      ptr_i16++;
                 }
            }
            break;     

  case DFNT_UINT32:
  
            ptr_u32 = (uint32 *)buffer;
            u32max = u32min = *ptr_u32;
            for(; ptr_u32 < (uint32 *)buffer + n_elements; ptr_u32++){
                 if(*ptr_u32 > u32max)u32max = *ptr_u32;
                 if(*ptr_u32 < u32min)u32min = *ptr_u32;
            }
            if(u32max == u32min){
                 for(i=0; i < n_elements; i++){
                      *pbuf = n_levels;
                      *pbuf++;
                 }
                 break;
            }
                 
            if(scaling == S_HISTO)
                 a1 =(float32)(N_BIN-1)/(u32max - u32min);
            else
                 a1 =(float32)n_levels/(u32max - u32min);
            b1 = (-a1*u32min) + 0.5;

            ptr_u32 = (uint32 *)buffer;
            if(scaling == S_HISTO){
                 for(i = 0; i< n_elements; i++){
                      *p2buf = (int16)(*ptr_u32*a1 + b1);
                      histo[*p2buf] ++;
                      p2buf ++;
                      ptr_u32 ++;
                 }
                 histo_eq(buffer, n_elements, 2, histo, N_BIN, n_levels);
            }
            else {
                 for(i = 0; i< n_elements; i++){
                      *pbuf = (unsigned char)(*ptr_u32*a1 + b1);
                      pbuf ++;
                      ptr_u32++;
                 }
            }
            break;     

  case DFNT_INT32:
  
            ptr_i32 = (int32 *)buffer;
            i32max = i32min = *ptr_i32;
            for(; ptr_i32 < (int32 *)buffer + n_elements; ptr_i32++){
                 if(*ptr_i32 > i32max)i32max = *ptr_i32;
                 if(*ptr_i32 < i32min)i32min = *ptr_i32;
            }
            if(i32max == i32min){
                 for(i=0; i< n_elements; i++){
                      *pbuf = n_levels;
                      *pbuf++;
                 }
                 break;
            }
            
            if(scaling == S_HISTO)     
                 a1 = (float32)(N_BIN-1)/(i32max - i32min);
            else
                 a1 = (float32)n_levels/(i32max - i32min);
            b1 = (-a1*i32min) + 0.5;

            ptr_i32 = (int32 *)buffer;
            if(scaling == S_HISTO){
                 for(i=0; i< n_elements; i++){
                      *p2buf = (int16)(*ptr_i32*a1 + b1);
                     histo[*p2buf] ++;
                        p2buf++;
                        ptr_i32 ++;
                 }
                 histo_eq(buffer, n_elements, 2, histo, N_BIN, n_levels);
            }
            else { 
                 for(i = 0; i< n_elements; i++){
                      *pbuf = (unsigned char)(*ptr_i32*a1 + b1);
                      pbuf ++;
                      ptr_i32++;
                 }
            }
          break;
          
  case DFNT_FLOAT32:
  
            ptr_f32 = (float32 *)buffer;
            f32max = f32min = *ptr_f32;
            for(; ptr_f32 < (float32 *)buffer + n_elements; ptr_f32++){
                 if(*ptr_f32 > f32max)f32max = *ptr_f32;
                 if(*ptr_f32 < f32min)f32min = *ptr_f32;
            }
            if(f32max == f32min){
                 for(i=0; i< n_elements; i++){
                      *pbuf = n_levels;
                      *pbuf++;
                 }
                 break;
            }
                 
            if(scaling == S_HISTO)
                 a1 = (N_BIN-1)/(f32max - f32min);
            else
                 a1 = n_levels/(f32max - f32min);
            b1 = (-a1*f32min) + 0.5;

            ptr_f32 = (float32 *)buffer;
            if(scaling == S_HISTO){
                 for(i=0; i< n_elements; i++){
                      *p2buf = (int16)(*ptr_f32*a1 + b1);
                      histo[*p2buf] ++;
                      p2buf ++;
                      ptr_f32 ++;
                 }
                 histo_eq(buffer, n_elements, 2, histo, N_BIN, n_levels);
            }
            else {     
                 for(i = 0; i< n_elements; i++){
                      *pbuf = (unsigned char)(*ptr_f32*a1 + b1);
                      pbuf ++;
                      ptr_f32++;
                 }
            }
            break;

  case DFNT_FLOAT64:
  
            ptr_f64 = (float64 *)buffer;
            f64max = f64min = *ptr_f64;
            for(; ptr_f64 < (float64 *)buffer + n_elements; ptr_f64++){
                 if(*ptr_f64 > f64max)f64max = *ptr_f64;
                 if(*ptr_f64 < f64min)f64min = *ptr_f64;
            }
            if(f64max == f64min){
                 for(i=0; i< n_elements; i++){
                      *pbuf = n_levels;
                      *pbuf++;
                 }
                 break;
            }
                 
            if(scaling == S_HISTO)
                 a = (N_BIN-1)/(f64max - f64min);
            else
                 a = n_levels/(f64max - f64min);                
            b = (-a * f64min) + 0.5;

            ptr_f64 = (float64 *)buffer;
            if(scaling == S_HISTO){
                 for(i=0; i< n_elements; i++){
                      *p2buf = (int16)(*ptr_f64*a + b);
                      histo[*p2buf]++;
                      p2buf++;
                      ptr_f64 ++;
                 }
                 histo_eq(buffer, n_elements, 2, histo, N_BIN, n_levels);
            }
            else {
                 for(i = 0; i< n_elements; i++){
                      *pbuf = (unsigned char)(*ptr_f64*a + b);
                      pbuf ++;
                      ptr_f64++;
                 }
            }
            break;
            
       default:
            return -1;
 }
 return 0;
}

void histo_eq(void *buffer, int n_elements, int type, long histo[], int n_bin, int n_levels)
{
     unsigned char *ptr;
     int16 *p2tr;
     int sum=0;
     int csum=0;
     unsigned char lookup[1024];
     int i;
     
     ptr = (unsigned char *)buffer;
     p2tr = (int16 *)buffer;

     for(i=0; i< n_bin; i++)
          sum +=histo[i];

     for(i=0; i< n_bin; i++){
          csum += histo[i];
          lookup[i] = (unsigned char)(csum/(double)sum * n_levels + 0.1);
/*          DBUG_PRINT(1,(LOGF,"csum =%d, look[%d]=%d, histo=%d\n", csum, i, lookup[i], histo[i]));
*/   }
     
     if(type == 1)
          for(; ptr < (unsigned char *)buffer + n_elements; ptr++)
               *ptr = lookup[*ptr];
     else
          for(; ptr < (unsigned char *)buffer + n_elements; ptr++, p2tr++)
               *ptr = lookup[*p2tr];
}
